"
 COPYRIGHT (c) 1988 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

"{ Package: 'stx:libbasic' }"

Magnitude subclass:#Character
	instanceVariableNames:'asciivalue'
	classVariableNames:'CharacterTable'
	poolDictionaries:''
	category:'Magnitude-General'
!

!Character class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1988 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

documentation
"
    This class represents characters.

    For now, unicode characters can be created and also put-into/taken from
    Unicode strings, but the attribute queries (such as isLetter, isDigit etc. are
    not yet unicode aware).

    Notice, that actual character objects are not used when characters
    are stored in strings, symbols or twoByteStrings; these only store
    a characters asciiValue for a more compact representation.
    The word 'asciiValue' is a historic leftover - actually, any integer
    code is allowed (i.e. characters are not limited to 8bit).

    Single byte Characters are unique; 
    i.e. for every asciiValue (0..255) there exists exactly one instance of 
    Character, which is shared (Character value:xxx checks for this, and returns 
    a reference to an existing instance).
    Other characters (i.e. asciivalue > 255) are not shared; i.e. these
    are created as required.

    This means: you may compare characters using #== ONLY IFF you are certain,
    that the characters ranges is 0..255. Otherwise, you HAVE TO compare
    using #=. (if in doubt, always compare using #=).
    Sorry for this inconvenience, but it is (practically) impossible to keep
    the possible maximum of 2^32 characters (Unicode) around, for that
    convenience alone.


    Methods marked as (JS) come from the manchester Character goody
    (CharacterComparing) by Jan Steinman, which allow Characters to be used as
    Interval elements (i.e. ($a to:$z) do:[...] ); 
    They are not a big deal, but convenient add-ons.
    Some of these have been modified a bit.

    WARNING: characters are known by compiler and runtime system -
             do not change the instance layout. 

    Also, although you can create subclasses of Character, the compiler always
    creates instances of Character for literals ...
    ... and other classes are hard-wired to always return instances of characters
    in some cases (i.e. String>>at:, Symbol>>at: etc.).
    Therefore, it may not make sense to create a character-subclass.

    [author:]
        Claus Gittinger

    [see also:]
        String TwoByteString
        StringCollection
"
! !

!Character class methodsFor:'instance creation'!

basicNew
    "catch new - Characters cannot be created with new"

    ^ self error:'Characters cannot be created with new'
!

digitValue:anInteger
    "return a character that corresponds to anInteger.
     0-9 map to $0-$9, 10-35 map to $A-$Z"

    |val "{ Class: SmallInteger }" |

    val := anInteger.
    (val between:0 and:9) ifTrue:[
	^ Character value:(val + ($0 asciiValue))
    ].
    (val between:10 and:35) ifTrue:[
	^ Character value:(val + ($A asciiValue - 10))
    ].
    ^self error:'value not in range 0 to 35'
!

utf8DecodeFrom:aStream
    "read and return a single unicode character from an UTF8 encoded stream"

    |fetchNext c1 c2 c3 c4 c5 ascii|

    c1 := aStream next.
    ascii := c1 asciiValue.
    ascii <= 16r7F ifTrue:[
        "/ 0xxxxxxx - 7 bits   
        ^ c1.
    ].

    (ascii bitAnd:2r11000000) == 2r10000000 ifTrue:[
        "/ out of sync (got an intermediate character)
        self error:'out of sync'.
        ^ c1.
    ].

    fetchNext := [  |ch|

                    ch := aStream next.
                    (ch asciiValue bitAnd:2r11000000) == 2r10000000 ifFalse:[
                        "/ followup chars must have 2r10 in high bits
                        self error:'bad encoding'.
                    ].
                    ch
                 ].

    (ascii bitAnd:2r11100000) == 2r11000000 ifTrue:[
        "/ 110xxxxx 10xxxxxx - 11 bits
        c2 := fetchNext value.
        ascii := c1 asciiValue bitAnd:16r1F.
        ascii := (ascii bitShift:6) bitOr:(c2 asciiValue bitAnd:16r3F).
        ascii <= 16r7F ifTrue:[
            self error:'invalid encoding'.
        ].
        ^ Character value:ascii        
    ].
    (ascii bitAnd:2r11110000) == 2r11100000 ifTrue:[
        "/ 1110xxxx 10xxxxxx 10xxxxxx - 16 bits
        c2 := fetchNext value.
        c3 := fetchNext value.
        ascii := c1 asciiValue bitAnd:16r0F.
        ascii := (ascii bitShift:6) bitOr:(c2 asciiValue bitAnd:16r3F).
        ascii := (ascii bitShift:6) bitOr:(c3 asciiValue bitAnd:16r3F).
        ascii <= 16r7FF ifTrue:[
            self error:'invalid encoding'.
        ].
        ^ Character value:ascii        
    ].

    "/ notice: currently, characters can only have 16bit encoding;
    "/ therefore the following will raise a runtime exception,

    (ascii bitAnd:2r11111000) == 2r11110000 ifTrue:[
        "/ 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx - 21 bits
        c2 := fetchNext value.
        c3 := fetchNext value.
        c4 := fetchNext value.
        ascii := c1 asciiValue bitAnd:16r07.
        ascii := (ascii bitShift:6) bitOr:(c2 asciiValue bitAnd:16r3F).
        ascii := (ascii bitShift:6) bitOr:(c3 asciiValue bitAnd:16r3F).
        ascii := (ascii bitShift:6) bitOr:(c4 asciiValue bitAnd:16r3F).
        ascii <= 16rFFFF ifTrue:[
            self error:'invalid encoding'.
        ].
        ^ Character value:ascii        
    ].

    (ascii bitAnd:2r11111100) == 2r11111000 ifTrue:[
        "/ 111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx - 26 bits
        c2 := fetchNext value.
        c3 := fetchNext value.
        c4 := fetchNext value.
        c5 := fetchNext value.
        ascii := c1 asciiValue bitAnd:16r03.
        ascii := (ascii bitShift:6) bitOr:(c2 asciiValue bitAnd:16r3F).
        ascii := (ascii bitShift:6) bitOr:(c3 asciiValue bitAnd:16r3F).
        ascii := (ascii bitShift:6) bitOr:(c4 asciiValue bitAnd:16r3F).
        ascii := (ascii bitShift:6) bitOr:(c5 asciiValue bitAnd:16r3F).
        ascii <= 16r1FFFFF ifTrue:[
            self error:'invalid encoding'.
        ].
        ^ Character value:ascii        
    ].

    (ascii bitAnd:2r11111110) == 2r11111100 ifTrue:[
        "/ 1111110x ... 10xxxxxx - any number of bits
        ascii := c1 asciiValue bitAnd:16r01.

        c2 := aStream peek.
        [ (c2 asciiValue bitAnd:2r11000000) == 2r10000000 ] whileTrue:[
            ascii := (ascii bitShift:6) bitOr:(c2 asciiValue bitAnd:16r3F).
            aStream next.
        ].
        ascii <= 16r3FFFFFF ifTrue:[
            self error:'invalid encoding'.
        ].
        ^ Character value:ascii        
    ].

    self error:'bad encoding'.
    ^ c1

    "
      Character utf8DecodeFrom:'a' readStream     
      Character utf8DecodeFrom:#[195 188] asString readStream   
    "                                            

    "test:

      |utf8Encoding original readBack|

      1 to:16rFFFF do:[:ascii |
        original := Character value:ascii.
        utf8Encoding := original asString utf8Encoded.
        readBack := Character utf8DecodeFrom:(utf8Encoding readStream).
        readBack asciiValue = ascii ifFalse:[
            self halt
        ]
      ]
    "
!

value:anInteger
    "return a character with asciivalue anInteger"

%{  /* NOCONTEXT */

    INT ascii;

    if (__isSmallInteger(anInteger)) {
        ascii = __intVal(anInteger);
        if ((unsigned)(ascii) <= MAX_IMMEDIATE_CHARACTER /* (ascii >= 0) && (ascii <= 255) */) {
            RETURN ( __MKCHARACTER(ascii) );
        }
    }
%}.
    (anInteger between:0 and:(CharacterTable size - 1)) ifTrue:[
        ^ CharacterTable at:(anInteger + 1)
    ].
    (anInteger between:16r100 and:16r3FFFFFFF) ifTrue:[
        ^ super basicNew setAsciiValue:anInteger
    ].
    "
     a characters ascii-code must be 0..16r3FFFFFFF.
     (i.e. only characters with up-to 30 bits are allowed.)
    "
    self error:'invalid ascii code for character'
! !

!Character class methodsFor:'Compatibility-ANSI'!

codePoint:asciiValue
    ^ self value:asciiValue
! !

!Character class methodsFor:'accessing untypeable characters'!

endOfInput
	"Answer the Character representing ctrl-d ."

	^self value: 4
!

leftParenthesis
	"Answer the Character representing a left parenthesis."

	^self value: 40
!

period
	"Answer the Character representing a carriage period."

	^self value: 46
!

poundSign
	"Answer the Character representing a pound sign."

	^self value: 35
!

rightParenthesis
	"Answer the Character representing a right parenthesis."

	^self value: 41
! !

!Character class methodsFor:'constants'!

backspace
    "return the backspace character"

    ^ Character value:8
!

bell
    "return the bell character"

    ^ Character value:7
!

cr
    "return the lineEnd character 
     - actually (in unix) this is a newline character"

    ^ Character value:10
!

del 
    "return the delete character"

    ^ Character value:16r7F 
!

doubleQuote
    "return the double-quote character"

    ^ Character value:34
!

esc
    "return the escape character"

    ^ Character value:27
!

euro
    "The Euro currency sign (notice: not all fonts support it).
     Thats the ascii-character code; the Unicode encoding is 0x8364"

    ^ Character value: 16r20AC

    "
     Transcript showCR:Character euro asString
    "
    "
     0 to:255 do:[:i |
        Transcript 
            show:'| '; show:((i printStringRadix:16) leftPaddedTo:2);
            show:' | '; show:(i printStringPaddedTo:3);
            show:' | '; show:(Character value:i);
            cr.
     ]
    "
!

excla
    "return the exclamation-mark character"
    ^ $!!
!

ff
    "return the form-feed character"

    ^ Character value:12
!

lf
    "return the newline/linefeed character"

    ^ Character value:10
!

linefeed
    "squeak compatibility: return the newline/linefeed character"

    ^ Character value:10
!

maxValue 
    "return the maximum asciiValue a character can have"

    ^ 16rFFFF
!

newPage
    "return the form-feed character"

    ^ Character value:12
!

nl
    "return the newline character"

    ^ Character value:10
!

null
    ^ Character value:0
!

quote
    "return the single-quote character"

    ^ Character value:39
!

return 
    "return the return character.
     In ST/X, this is different from cr - for Unix reasons."

    ^ Character value:13 
!

space
    "return the blank character"

    ^ Character value:32
!

tab
    "return the tabulator character"

    ^ Character value:9
! !

!Character class methodsFor:'primitive input'!

fromUser
    "return a character from the keyboard (C's standard input stream)
     - this should only be used for emergency evaluators and the like."

%{  /* NOCONTEXT */
    int c;

    c = getchar();
    if (c < 0) {
        RETURN (nil);
    }
    RETURN ( __MKCHARACTER(c & 0xFF) );
%}.
    ^ Stdin next
! !

!Character class methodsFor:'queries'!

allCharacters
    "added for squeak compatibility: return a collection of all chars"

    ^ CharacterTable
        
    "
     Character allCharacters 
    "
!

hasSharedInstances
    "return true if this class has shared instances, that is, instances
     with the same value are identical.
     False is returned here, only redefined in classes which have unified
     instances (or should be treated so)."

    ^ true


!

isBuiltInClass
    "return true if this class is known by the run-time-system.
     Here, true is returned for myself, false for subclasses."

    ^ self == Character

    "Modified: 23.4.1996 / 15:56:39 / cg"
!

separators
    "added for squeak compatibility: return a collection of separator chars"

    ^ Array 
        with:Character space
        with:Character return
        "/ with:Character cr
        with:Character tab
        with:Character lf
        with:Character ff
        
    "
     Character separators
    "
! !

!Character methodsFor:'Compatibility-ANSI'!

codePoint
    "return the codePoint of myself.
     Traditionally, this was named 'asciiValue'; 
     however, characters are not limited to 8bit characters."

    ^ asciivalue
! !

!Character methodsFor:'Compatibility-Dolphin'!

isAlphaNumeric
    "return true, if I am a letter or a digit
     Please use isLetterOrDigit for compatibility reasons (which is ANSI)."

    ^ self isLetterOrDigit
!

isAlphabetic
    "return true, if I am a letter.
     Please use isLetter for compatibility reasons (which is ANSI)."

    ^ self isLetter
!

isControl
    "return true if I am a control character (i.e. ascii value < 32)"

    ^ self isControlCharacter
!

isHexDigit
    "return true if I am a valid hexadecimal digit"

    ^ '0123456789abcdefABCDEF' includes:self

    "
     $a isHexDigit
    "
!

isPunctuation
    "the code below is not unicode aware"

    ^ (asciivalue between:16r21 and:16r40)
      or:[ (asciivalue between:16r5B and:16r60)
      or:[ (asciivalue between:123 and:126)
      or:[ (asciivalue between:161 and:191)
      or:[ (asciivalue == 215 )
      or:[ (asciivalue == 247 ) ]]]]]
! !

!Character methodsFor:'accessing'!

asciiValue
    "return the asciivalue of myself.
     The name 'asciiValue' is a historic leftover:
        characters are not limited to 8bit characters.
     So the actual value returned is a codePoint (i.e. full potential for 31bit encoding).
     PP has removed this method with 4.1 and providing asInteger instead."

    ^ asciivalue

    "Modified: 27.6.1996 / 12:34:34 / cg"
!

instVarAt:index put:anObject
    "catch instvar access - asciivalue may not be changed"

    self error:'Characters may not be modified'
!

newDigitValue
	"Answer 0-9 if the receiver is $0-$9, 10-35 if it is $A-$Z, and < 0 otherwise. 
	This is used to parse literal numbers of radix 2-36."

	| code value |
	code := self asInteger.
	value := code - $0 asInteger.
	value <= 9 ifTrue: [^value].
	value := code - $A asInteger.
	(value >= 0 and: [value < 26])
		ifTrue: [^value + 10].
	^-1
! !

!Character methodsFor:'arithmetic'!

+ aMagnitude
    "Return the Character that is <aMagnitude> higher than the receiver. 
     Wrap if the resulting value is not a legal Character value. (JS)"

    ^ Character value:(asciivalue + aMagnitude asInteger \\ 16rFFFF)

    "
     $A + 5
    "

    "Modified: 27.6.1996 / 12:34:51 / cg"
!

- aMagnitude
    "Return the Character that is <aMagnitude> lower than the receiver.  
     Wrap if the resulting value is not a legal Character value. (JS)
     claus: 
        modified to return the difference as integer, if the argument
        is another character. If the argument is a number, a character is
        returned."

    aMagnitude isCharacter ifTrue:[
        ^ self asciiValue - aMagnitude asciiValue
    ].
    ^ Character value:(asciivalue - aMagnitude asInteger \\ 16rFFFF)

    "
     $z - $a  
     $d - 3
    "

    "Modified: 27.6.1996 / 12:35:34 / cg"
!

// aMagnitude
    "Return the Character who's value is the receiver divided by <aMagnitude>. 
     Wrap if the resulting value is not a legal Character value. (JS)"

    ^ Character value:(asciivalue // aMagnitude asInteger \\ 16rFFFF)
!

\\ aMagnitude
    "Return the Character who's value is the receiver modulo <aMagnitude>.  
     Wrap if the resulting value is not a legal Character value. (JS)"

    ^ Character value:(asciivalue \\ aMagnitude asInteger \\ 16rFFFF)
! !

!Character methodsFor:'binary storage'!

hasSpecialBinaryRepresentation
    "return true, if the receiver has a special binary representation"

    ^ true
!

storeBinaryOn:stream manager:manager
    "store a binary representation of the receiver on stream;
     redefined, since single-byte characters are stored more compact
     with a special type-code followed by the asciiValue."

    (asciivalue < 256) ifTrue:[
	stream nextPut:manager codeForCharacter; nextPut:asciivalue.
    ] ifFalse:[
	stream nextPut:manager codeForTwoByteCharacter.
	stream nextPutShort:asciivalue MSB:true
    ]
! !

!Character methodsFor:'comparing'!

< aCharacter
    "return true, if the arguments asciiValue is greater than the receiver's"

    ^ (asciivalue < aCharacter asInteger)
!

<= aCharacter
    "return true, if the arguments asciiValue is greater or equal to the receiver's"

    ^ (asciivalue <= aCharacter asInteger)
!

= aCharacter
    "return true, if the argument, aCharacter is the same character
     Redefined to take care of 16bit characters."

    self == aCharacter ifTrue:[^ true].
    aCharacter isCharacter ifFalse:[^ false].
    ^ (asciivalue = aCharacter asciiValue)
!

> aCharacter
    "return true, if the arguments asciiValue is less than the receiver's"

    ^ (asciivalue > aCharacter asInteger)
!

>= aCharacter
    "return true, if the arguments asciiValue is less or equal to the receiver's"

    ^ (asciivalue >= aCharacter asInteger)
!

hash
    "return an integer useful for hashing"

    ^ asciivalue
!

identityHash
    "return an integer useful for hashing on identity"

    asciivalue <= 255 ifTrue:[
	^ 4096 + asciivalue
    ].
    ^ super identityHash
!

sameAs:aCharacter
    "return true, if the argument, aCharacter is the same character,
     ignoring case differences."

    self == aCharacter ifTrue:[^ true].
    ^ self asLowercase = aCharacter asLowercase
!

~= aCharacter
    "return true, if the argument, aCharacter is not the same character
     Redefined to take care of 16bit characters."

    self == aCharacter ifTrue:[^ false].
    aCharacter isCharacter ifFalse:[^ true].
    ^ (asciivalue ~~ aCharacter asciiValue)
! !

!Character methodsFor:'converting'!

asCharacter
    "usually sent to integers, but redefined here to allow integers
     and characters to be used commonly without a need for a test."

    ^ self

    "
     32 asCharacter  
    "
!

asInteger
    "return an Integer with my ascii-value.
     OWST4.2 compatibility (sigh)"

    ^ asciivalue
!

asLowercase
    "return a character with same letter as the receiver,
     but lowercase (the receiver if its lowercase or nonLetter)"

    |code "{Class: SmallInteger }"|

    code := asciivalue.

    code <= 16r00FF ifTrue:[
        "/ ISO Latin-1
        ((code >= $A asciiValue) and:[code <= $Z asciiValue]) ifTrue:[
            ^ Character value:(code + ($a asciiValue - $A asciiValue))
        ].
        code < 16r00C0 ifTrue:[^ self].
        code >= 16r00DF ifTrue:[^ self].
        code == 16r00D7 ifTrue:[^ self].
        ^ Character value:(code + 16r20)
    ].

    self isLowercase ifTrue:[^ self].
    self isNationalLetter ifFalse:[^ self].

    code <= 16r01FF ifTrue:[
        ^ Character value:(code + 1)
    ].

    code <= 16r02FF ifTrue:[
        code <= 16r021F ifTrue:[
            code even ifTrue:[ ^ Character value:(code + 1) ].
            ^ self.
        ].
        ^ self.
    ].

    code <= 16r03FF ifTrue:[
        (code between:16r0391 and:16r03AA) ifTrue:[ ^ Character value:(code + 32)].
        ^ self.
    ].

    code <= 16r04FF ifTrue:[
        (code between:16r0400 and:16r040F) ifTrue:[ ^ Character value:(code + 16r50)].
        (code between:16r0410 and:16r042F) ifTrue:[ ^ Character value:(code + 16r20)].
        "/ (code between:16r0460 and:16r0480) ifTrue:[ ^ Character value:(code + 1)].
        ^ Character value:(code + 1)
    ].

    code <= 16r05FF ifTrue:[
    ].

    code <= 16r06FF ifTrue:[
    ].

    code <= 16rFFFF ifTrue:[
        (code between:16rFF21 and:16rFF3A) ifTrue:[ ^ Character value:(code + 16r20)].
    ].

    ^ self

    "
     $A asLowercase 
     $1 asLowercase  
     (Character value:16r10d) asLowercase  
     (Character value:16r10d) asUppercase  
    "

    "Modified: 10.1.1997 / 15:35:44 / cg"
!

asString
    "return a string of len 1 with myself as contents"

%{  /* NOCONTEXT */
    char buffer[2];
    OBJ s;
    unsigned val;

    val = __intVal(_characterVal(self));
    if (val <= 0xFF) {
        buffer[0] = (char) val;
        buffer[1] = '\0';
        s = __MKSTRING_L(buffer, 1);
        if (s != nil) {
            RETURN (s);
        }
    }
    if (val <= 0xFFFF) {
        s = __MKEMPTYUSTRING(1);
        if (s != nil) {
            __Unicode16StringInstPtr(s)->s_element[0] = val;
            RETURN (s);
        }
    }
%}.
    asciivalue > 16rFF ifTrue:[
        asciivalue > 16rFFFF ifTrue:[
            ^ (Unicode32String new:1) at:1 put:self; yourself
        ].
        ^ (Unicode16String new:1) at:1 put:self; yourself
    ].

    ^ (String new:1) at:1 put:self; yourself.

    "
     (Character value:16rB5) asString     
     (Character value:16r1B5) asString     
    "
!

asSymbol
    "return a unique symbol which prints like I print"

    ^ Symbol internCharacter:self
!

asUnicodeString
    "return a unicode string of len 1 with myself as contents.
     This will vanish, as we now (rel5.2.x) use Unicode as default."

    asciivalue > 16rFFFF ifTrue:[
        ^ (Unicode32String new:1) at:1 put:self; yourself.
    ].
    ^ (Unicode16String new:1) at:1 put:self; yourself.
!

asUppercase
    "return a character with same letter as the receiver,
     but uppercase (the receiver if its uppercase or nonLetter)"

    |code "{Class: SmallInteger }"|

"/ the old code:
"/    self isLowercase ifFalse:[^ self].
"/    ^ Character value:(asciivalue - 32)

    code := asciivalue.
    code <= 16r00FF ifTrue:[
        "/ ISO Latin-1
        ((code >= $a asciiValue) and:[code <= $z asciiValue]) ifTrue:[
            ^ Character value:(code + ($A asciiValue - $a asciiValue))
        ].
        code < 16r00E0 ifTrue:[
            code == 16r00B5 ifTrue:[^ Character value:16r039C].  "/ greek mu - Capital not in ISO-latin 1
            ^ self
        ].

        code == 16r00F7 ifTrue:[^ self]. "/ division
        code == 16r00FF ifTrue:[^ Character value:16r0178].  "/ y diaresis - Capital not in ISO-latin 1
        code == 16r00FF ifTrue:[^ self]. 
        ^ Character value:(code - 16r20)
    ].
    self isUppercase ifTrue:[^ self].
    self isNationalLetter ifFalse:[^ self].

    code <= 16r01FF ifTrue:[
        (code == 16r131) ifTrue:[ ^ $I ].   "/ dotless i
        (code == 16r138) ifTrue:[ ^ self ]. "/ greenlandic kra
        (code == 16r149) ifTrue:[ ^ self ]. "/ 
        (code == 16r17F) ifTrue:[ ^ $S ].   "/ small fractur/gaelic s
        (code == 16r180) ifTrue:[ ^ self ]. "/ small b with stroke
        (code == 16r181) ifTrue:[ ^ self ]. "/ upper B with hook
        ^ Character value:(code - 1)
    ].
    code <= 16r02FF ifTrue:[
        code <= 16r021F ifTrue:[
            code even ifTrue:[ ^ self ].
            ^ Character value:(code - 1).
        ].
        ^ self.
    ].

    code <= 16r03FF ifTrue:[
        code >= 16r03B1 ifTrue:[
            code <= 16r03CA ifTrue:[ ^ Character value:(code - 32)].
        ].
        ^ self.
    ].

    code <= 16r04FF ifTrue:[
        (code between:16r0450 and:16r045F) ifTrue:[ ^ Character value:(code - 16r50)].
        (code between:16r0430 and:16r044F) ifTrue:[ ^ Character value:(code - 16r20)].
        "/ (code between:16r0461 and:16r0481) ifTrue:[ ^ Character value:(code - 1)].
        ^ Character value:(code - 1)
    ].

    code <= 16r05FF ifTrue:[
    ].

    code <= 16rFFFF ifTrue:[
        (code between:16rFF41 and:16rFF5A) ifTrue:[ ^ Character value:(code - 16r20)].
    ].

    ^ self

    "
     $A asLowercase 
     $1 asLowercase  
    "

    "Modified: 10.1.1997 / 15:35:51 / cg"
!

decodeAsLiteralArray
    "given a literalEncoding in the receiver,
     create & return the corresponding object.
     The inverse operation to #literalArrayEncoding."

    ^ self

    "Created: / 27.10.1997 / 14:40:16 / cg"
!

digitValue
    "return my digitValue for any base"

    |ascii "{ Class: SmallInteger }" |

    ascii := asciivalue.
    (ascii between:($0 asciiValue) and:($9 asciiValue)) ifTrue:[
	^ ascii - $0 asciiValue
    ].
    (ascii between:($a asciiValue) and:($z asciiValue)) ifTrue:[
	^ ascii + (10 - $a asciiValue)
    ]. 
    (ascii between:($A asciiValue) and:($Z asciiValue)) ifTrue:[
	^ ascii + (10 - $A asciiValue)
    ]. 

"remove error below for X3J20 conformance ... "
    self error:'bad character'.
" "
    ^ -1
!

literalArrayEncoding
    "encode myself as an array literal, from which a copy of the receiver
     can be reconstructed with #decodeAsLiteralArray."

    ^ self

    "Created: / 27.10.1997 / 14:40:37 / cg"
!

to:aMagnitude
    "Return an Interval over the characters from the receiver to <aMagnitude>. 
     Wrap <aMagnitude> if it is not a legal Character value. (JS)"

    ^ Interval from:self to:(aMagnitude \\ 256)
!

u_asUppercase
%{
    static const unsigned short
    __toUpper[] = {
    /* From    To             Every   Diff   */ 
       0x0061, ((0x19 << 8) | 0x01), 0xffe0  ,
       0x00b5, ((0x00 << 8) | 0x3b), 0x02e7  ,
       0x00e0, ((0x16 << 8) | 0x01), 0xffe0  ,
       0x00f8, ((0x06 << 8) | 0x01), 0xffe0  ,
       0x00ff, ((0x00 << 8) | 0x01), 0x0079  ,
       0x0101, ((0x2e << 8) | 0x02), 0xffff  ,
       0x0131, ((0x00 << 8) | 0x02), 0xff18  ,
       0x0133, ((0x04 << 8) | 0x02), 0xffff  ,
       0x013a, ((0x0e << 8) | 0x02), 0xffff  ,
       0x014b, ((0x2c << 8) | 0x02), 0xffff  ,
       0x017a, ((0x04 << 8) | 0x02), 0xffff  ,
       0x017f, ((0x00 << 8) | 0x01), 0xfed4  ,
       0x0183, ((0x02 << 8) | 0x02), 0xffff  ,
       0x0188, ((0x04 << 8) | 0x04), 0xffff  ,
       0x0192, ((0x00 << 8) | 0x06), 0xffff  ,
       0x0195, ((0x00 << 8) | 0x03), 0x0061  ,
       0x0199, ((0x08 << 8) | 0x08), 0xffff  ,
       0x01a3, ((0x02 << 8) | 0x02), 0xffff  ,
       0x01a8, ((0x05 << 8) | 0x05), 0xffff  ,
       0x01b0, ((0x04 << 8) | 0x04), 0xffff  ,
       0x01b6, ((0x03 << 8) | 0x03), 0xffff  ,
       0x01bd, ((0x00 << 8) | 0x04), 0xffff  ,
       0x01bf, ((0x00 << 8) | 0x02), 0x0038  ,
       0x01c5, ((0x00 << 8) | 0x06), 0xffff  ,
       0x01c6, ((0x00 << 8) | 0x01), 0xfffe  ,
       0x01c8, ((0x00 << 8) | 0x02), 0xffff  ,
       0x01c9, ((0x00 << 8) | 0x01), 0xfffe  ,
       0x01cb, ((0x00 << 8) | 0x02), 0xffff  ,
       0x01cc, ((0x00 << 8) | 0x01), 0xfffe  ,
       0x01ce, ((0x0e << 8) | 0x02), 0xffff  ,
       0x01dd, ((0x00 << 8) | 0x01), 0xffb1  ,
       0x01df, ((0x10 << 8) | 0x02), 0xffff  ,
       0x01f2, ((0x00 << 8) | 0x03), 0xffff  ,
       0x01f3, ((0x00 << 8) | 0x01), 0xfffe  ,
       0x01f5, ((0x04 << 8) | 0x04), 0xffff  ,
       0x01fb, ((0x24 << 8) | 0x02), 0xffff  ,
       0x0223, ((0x10 << 8) | 0x02), 0xffff  ,
       0x0253, ((0x00 << 8) | 0x20), 0xff2e  ,
       0x0254, ((0x00 << 8) | 0x01), 0xff32  ,
       0x0256, ((0x01 << 8) | 0x01), 0xff33  ,
       0x0259, ((0x00 << 8) | 0x02), 0xff36  ,
       0x025b, ((0x00 << 8) | 0x02), 0xff35  ,
       0x0260, ((0x00 << 8) | 0x05), 0xff33  ,
       0x0263, ((0x00 << 8) | 0x03), 0xff31  ,
       0x0268, ((0x00 << 8) | 0x05), 0xff2f  ,
       0x0269, ((0x06 << 8) | 0x06), 0xff2d  ,
       0x0272, ((0x00 << 8) | 0x03), 0xff2b  ,
       0x0275, ((0x00 << 8) | 0x03), 0xff2a  ,
       0x0280, ((0x03 << 8) | 0x03), 0xff26  ,
       0x0288, ((0x00 << 8) | 0x05), 0xff26  ,
       0x028a, ((0x01 << 8) | 0x01), 0xff27  ,
       0x0292, ((0x00 << 8) | 0x07), 0xff25  ,
       0x0345, ((0x00 << 8) | 0xb3), 0x0054  ,
       0x03ac, ((0x00 << 8) | 0x67), 0xffda  ,
       0x03ad, ((0x02 << 8) | 0x01), 0xffdb  ,
       0x03b1, ((0x10 << 8) | 0x01), 0xffe0  ,
       0x03c2, ((0x00 << 8) | 0x01), 0xffe1  ,
       0x03c3, ((0x08 << 8) | 0x01), 0xffe0  ,
       0x03cc, ((0x00 << 8) | 0x01), 0xffc0  ,
       0x03cd, ((0x01 << 8) | 0x01), 0xffc1  ,
       0x03d0, ((0x00 << 8) | 0x02), 0xffc2  ,
       0x03d1, ((0x00 << 8) | 0x01), 0xffc7  ,
       0x03d5, ((0x00 << 8) | 0x04), 0xffd1  ,
       0x03d6, ((0x00 << 8) | 0x01), 0xffca  ,
       0x03db, ((0x14 << 8) | 0x02), 0xffff  ,
       0x03f0, ((0x00 << 8) | 0x01), 0xffaa  ,
       0x03f1, ((0x00 << 8) | 0x01), 0xffb0  ,
       0x03f2, ((0x00 << 8) | 0x01), 0xffb1  ,
       0x0430, ((0x1f << 8) | 0x01), 0xffe0  ,
       0x0450, ((0x0f << 8) | 0x01), 0xffb0  ,
       0x0461, ((0x20 << 8) | 0x02), 0xffff  ,
       0x048d, ((0x32 << 8) | 0x02), 0xffff  ,
       0x04c2, ((0x02 << 8) | 0x02), 0xffff  ,
       0x04c8, ((0x04 << 8) | 0x04), 0xffff  ,
       0x04d1, ((0x24 << 8) | 0x02), 0xffff  ,
       0x04f9, ((0x00 << 8) | 0x04), 0xffff  ,
       0x0561, ((0x25 << 8) | 0x01), 0xffd0  ,
       0x1e01, ((0x94 << 8) | 0x02), 0xffff  ,
       0x1e9b, ((0x00 << 8) | 0x06), 0xffc5  ,
       0x1ea1, ((0x58 << 8) | 0x02), 0xffff  ,
       0x1f00, ((0x07 << 8) | 0x01), 0x0008  ,
       0x1f10, ((0x05 << 8) | 0x01), 0x0008  ,
       0x1f20, ((0x07 << 8) | 0x01), 0x0008  ,
       0x1f30, ((0x07 << 8) | 0x01), 0x0008  ,
       0x1f40, ((0x05 << 8) | 0x01), 0x0008  ,
       0x1f51, ((0x06 << 8) | 0x02), 0x0008  ,
       0x1f60, ((0x07 << 8) | 0x01), 0x0008  ,
       0x1f70, ((0x01 << 8) | 0x01), 0x004a  ,
       0x1f72, ((0x03 << 8) | 0x01), 0x0056  ,
       0x1f76, ((0x01 << 8) | 0x01), 0x0064  ,
       0x1f78, ((0x01 << 8) | 0x01), 0x0080  ,
       0x1f7a, ((0x01 << 8) | 0x01), 0x0070  ,
       0x1f7c, ((0x01 << 8) | 0x01), 0x007e  ,
       0x1f80, ((0x07 << 8) | 0x01), 0x0008  ,
       0x1f90, ((0x07 << 8) | 0x01), 0x0008  ,
       0x1fa0, ((0x07 << 8) | 0x01), 0x0008  ,
       0x1fb0, ((0x01 << 8) | 0x01), 0x0008  ,
       0x1fb3, ((0x00 << 8) | 0x02), 0x0009  ,
       0x1fbe, ((0x00 << 8) | 0x0b), 0xe3db  ,
       0x1fc3, ((0x00 << 8) | 0x05), 0x0009  ,
       0x1fd0, ((0x01 << 8) | 0x01), 0x0008  ,
       0x1fe0, ((0x01 << 8) | 0x01), 0x0008  ,
       0x1fe5, ((0x00 << 8) | 0x04), 0x0007  ,
       0x1ff3, ((0x00 << 8) | 0x0e), 0x0009  ,
       0x2170, ((0x0f << 8) | 0x01), 0xfff0  ,
       0x24d0, ((0x19 << 8) | 0x01), 0xffe6  ,
       0xff41, ((0x19 << 8) | 0x01), 0xffe0   
    };

    REGISTER unsigned __codePoint;
    REGISTER short *__p;

    __codePoint = __intVal(__INST(asciivalue));
    for (__p = __toUpper; __p < ((char *)__toUpper) + sizeof(__toUpper); __p += 3) {
        short rangeStart, rangeSize, rangeEnd, mod;

        rangeStart = __p[0];
        if (__codePoint < rangeStart) break;

        rangeSize = __p[1] >> 8;
        rangeEnd = rangeStart + rangeSize;
        if (__codePoint <= rangeEnd) {
            mod = __p[1] & 0xFF;       
            if (((__codePoint - rangeStart) % mod) == 0) {
                OBJ newChar;
                unsigned short newCodePoint;

                newCodePoint = __codePoint + __p[2];
                if (newCodePoint <= 0xFF) {
                    RETURN (__MKCHARACTER(newCodePoint)) ;
                }
                RETURN (__MKUCHARACTER(newCodePoint)) ;
            }
            RETURN (self);
        }
    }
    RETURN (self);
getOutOfHere: ;
%}.
    ^ ObjectMemory allocationFailureSignal raise.

    "
     $A asLowercase 
     $a u_asUppercase 
     $1 asLowercase  
     (Character value:16rB5) u_asUppercase  

     0 to:16r05FF do:[:a |
        |c cU1 cU2|

        c := Character value:a.
        cU1 := c asUppercase.
        cU2 := c u_asUppercase.
        self assert:(cU1 = cU2).
     ]
    "    

    "Modified: 10.1.1997 / 15:35:51 / cg"
!

utf8Encoded
    "convert a character from ISO8859-1 to its UTF-8 encoding.
     this returns a String"

    |s|

    asciivalue <= 16r7F ifTrue:[
        ^ self asString.
    ].

    s := (String new:6) writeStream.
    self utf8EncodedOn:s.
    ^ s contents
!

utf8EncodedOn:aStream
    "append my UTF-8 representation to the argument, aStream.
     Up to 31 bits can be encoded in up to 6 bytes. 
     However, currently, characters are limited to 16 bits."

    |b1 b2 b3 b4 b5 v|

    asciivalue <= 16r7F ifTrue:[
        aStream nextPut:self.
        ^ self.
    ].

    b1 := Character value:((asciivalue bitAnd:16r3F) bitOr:2r10000000).
    v := asciivalue bitShift:-6.
    v <= 16r1F ifTrue:[
        aStream nextPut:(Character value:(v bitOr:2r11000000)).
        aStream nextPut:b1.
        ^ self.
    ].
    b2 := Character value:((v bitAnd:16r3F) bitOr:2r10000000).
    v := v bitShift:-6.
    v <= 16r0F ifTrue:[
        aStream nextPut:(Character value:(v bitOr:2r11100000)).
        aStream nextPut:b2; nextPut:b1.
        ^ self.
    ].
    b3 := Character value:((v bitAnd:16r3F) bitOr:2r10000000).
    v := v bitShift:-6.
    v <= 16r07 ifTrue:[
        aStream nextPut:(Character value:(v bitOr:2r11110000)).
        aStream nextPut:b3; nextPut:b2; nextPut:b1.
        ^ self.
    ].
    b4 := Character value:((v bitAnd:16r3F) bitOr:2r10000000).
    v := v bitShift:-6.
    v <= 16r03 ifTrue:[
        aStream nextPut:(Character value:(v bitOr:2r11111000)).
        aStream nextPut:b4; nextPut:b3; nextPut:b2; nextPut:b1.
        ^ self.
    ].
    b5 := Character value:((v bitAnd:16r3F) bitOr:2r10000000).
    v := v bitShift:-6.
    v <= 16r01 ifTrue:[
        aStream nextPut:(Character value:(v bitOr:2r11111100)).
        aStream nextPut:b5; nextPut:b4; nextPut:b3; nextPut:b2; nextPut:b1.
        ^ self.
    ].
    "/ cannot happen - we only support up to 16 bit characters
    self error:'ascii value > 31bit in utf8Encode'.
        
    "
      $a utf8Encoded   
      $ utf8Encoded asByteArray  
      (Character value:16r1fff) utf8Encoded asByteArray      
      (Character value:16rffff) utf8Encoded asByteArray      
      (Character value:16r800) utf8Encoded asByteArray      
    "
! !

!Character methodsFor:'copying'!

copy
    "return a copy of myself
     reimplemented since characters are unique"

     ^ self
!

copyUpToLast: char 

	^self
!

deepCopyUsing:aDictionary
    "return a deep copy of myself
     reimplemented since characters are immutable"

     ^ self
!

shallowCopy
    "return a shallow copy of myself
     reimplemented since characters are immutable"

     ^ self
!

simpleDeepCopy
    "return a deep copy of myself
     reimplemented since characters are immutable"

     ^ self
! !

!Character methodsFor:'encoding'!

encodeOn:anEncoder with:aParameter

    anEncoder encodeCharacter:self with:aParameter


!

rot13
     "Usenet: from `rotate alphabet 13 places'] 
      The simple Caesar-cypher encryption that replaces each English
      letter with the one 13 places forward or back along the alphabet, 
      so that 'The butler did it!!' becomes 'Gur ohgyre qvq vg!!'
      Most Usenet news reading and posting programs include a rot13 feature. 
      It is used to enclose the text in a sealed wrapper that the reader must choose
      to open -- e.g., for posting things that might offend some readers, or spoilers. 
      A major advantage of rot13 over rot(N) for other N is that it
      is self-inverse, so the same code can be used for encoding and decoding."

    (((self >= $a) and:[ self < $n]) or:[(self >= $A) and:[ self < $N]])
        ifTrue: [ ^ Character value:(self asciiValue + 13) ].
    (((self > $m) and:[ self <= $z]) or:[(self > $M) and:[ self <= $Z]])
        ifTrue: [ ^ Character value:(self asciiValue - 13) ].
    ^ self

    "
     $h rot13      
     $h rot13 rot13           
     'The butler did it!!' rot13             -> 'Gur ohgyre qvq vg!!'
     'The butler did it!!' rot13 rot13       -> 'The butler did it!!'
    "
! !

!Character methodsFor:'national testing'!

isNationalAlphaNumeric
    "return true, if the receiver is a letter or digit in the
     current language (Language variable)"

    self isNationalLetter ifTrue:[^ true].
    ^ self isDigit 
!

isNationalDigit
    "return true, if the receiver is a digit.
     This assumes unicode encoding.
     WARNING: this method is not complete."

    |codePoint "{ Class SmallInteger }"|

    codePoint := asciivalue.

    codePoint <= 16rFF ifTrue:[                "/ u00xx - unicode latin1 page 
        (codePoint between:($0 asciiValue) and:($9 asciiValue)) ifTrue:[^ true].
        ^ false
    ].

    (codePoint between:16rFF10 and:16rFF19) ifTrue:[ ^ true].
    ^ false.
!

isNationalLetter
    "return true, if the receiver is a letter.
     This assumes unicode encoding.
     WARNING: this method is not complete."

    |codePoint "{ Class SmallInteger }"|

    codePoint := asciivalue.

    codePoint <= 16rFF ifTrue:[                "/ u00xx - unicode latin1 page 
        (codePoint between:($a asciiValue) and:($z asciiValue)) ifTrue:[^ true].
        (codePoint between:($A asciiValue) and:($Z asciiValue)) ifTrue:[^ true].
        codePoint < 16rC0 ifTrue:[^ false].
        codePoint == 16rD7 ifTrue:[^ false].
        codePoint == 16rF7 ifTrue:[^ false].
        ^ true
    ].
    codePoint <= 16r1FF ifTrue:[               "/ u0100..u017F - unicode european latin (Latin ext-A) page
        ^ true                                 "/ u0180..u01FF - unicode european latin (Latin ext-B) page
    ].
    codePoint <= 16r2FF ifTrue:[               "/ u02xx - unicode latin page
        codePoint <= 16r2B8 ifTrue:[               
            ^ true
        ].
        ^ false
    ].                                         
    codePoint <= 16r3FF ifTrue:[               "/ u03xx - unicode accent & greek page
        codePoint <= 16r385 ifTrue:[ ^ false ].
        codePoint == 16r387 ifTrue:[ ^ false ].
        ^ true
    ].
    codePoint <= 16r4FF ifTrue:[               "/ u04xx - unicode cyrillic
        ^ true
    ].
    codePoint <= 16r5FF ifTrue:[               "/ u05xx - 
        codePoint <= 16r50F ifTrue:[           "/ u05xx - unicode cyrillic supplementary
            ^ true
        ].
        codePoint <= 16r530 ifTrue:[ ^ false ]. 
        codePoint <= 16r556 ifTrue:[           "/ u05xx - unicode armenian
            ^ true
        ].
        codePoint <= 16r550 ifTrue:[ ^ false ]. 
        codePoint <= 16r587 ifTrue:[           
            ^ true
        ].
        codePoint <= 16r5CF ifTrue:[ ^ false ]."/ u05xx - unicode hebrew           
        codePoint <= 16r5F2 ifTrue:[          
            ^ true
        ].
        ^ false.
    ].
    codePoint <= 16r6FF ifTrue:[               "/ u06xx - arabic
        codePoint <= 16r620 ifTrue:[ ^ false ]. 
        codePoint <= 16r64A ifTrue:[ ^ true ]. 
        codePoint <= 16r66D ifTrue:[ ^ false ]. 
        codePoint <= 16r6D3 ifTrue:[ ^ true ]. 
        codePoint == 16r6D5 ifTrue:[ ^ true ]. 
        codePoint == 16r6EE ifTrue:[ ^ true ]. 
        codePoint == 16r6EF ifTrue:[ ^ true ]. 
        ^ false.
    ].
    codePoint <= 16r7FF ifTrue:[               "/ u06xx - syriac / thaana
        codePoint < 16r710 ifTrue:[ ^ false ]. 
        codePoint < 16r74F ifTrue:[ ^ true ]. 
        codePoint < 16r780 ifTrue:[ ^ false ]. 
        codePoint < 16r7B0 ifTrue:[ ^ true ]. 
        ^ false.
    ].
    codePoint <= 16rDFF ifTrue:[               
        "/ self halt:'unimplemented character range'.
        ^ false.
    ].
    codePoint <= 16rEFF ifTrue:[               "/ u0Exx - thai / lao
        codePoint < 16rE50 ifTrue:[ ^ true].
        codePoint < 16rE80 ifTrue:[ ^ false].
        codePoint < 16rED0 ifTrue:[ ^ true].
        ^ false.
    ].
    codePoint <= 16rFEFF ifTrue:[               
        ^ false.
    ].
    codePoint <= 16rFFFF ifTrue:[               
        (codePoint between:16rFF21 and:16rFF3A) ifTrue:[ ^ true].
        (codePoint between:16rFF41 and:16rFF5A) ifTrue:[ ^ true].
        (codePoint between:16rFF66 and:16rFF9D) ifTrue:[ ^ true].   "/ HalfWidth-Katakana
        (codePoint between:16rFFA0 and:16rFFDC) ifTrue:[ ^ true].   "/ HalfWidth-Hangul
        ^ false.
    ].
    ^ false.

    "Modified: 21.1.1997 / 11:13:02 / cg"
! !

!Character methodsFor:'printing & storing'!

displayString
    "return a string used when the receiver is to be displayed
     in an inspector kind-of-thing"

    ^ (self storeString , ' "16r' , (asciivalue printStringRadix:16) , '"')
!

isLiteral
    "return true, if the receiver can be used as a literal constant in ST syntax
     (i.e. can be used in constant arrays)"

    ^ true
!

print
    "print myself on stdout.
     This method does NOT (by purpose) use the stream classes and 
     will therefore work even in case of emergency (but only, if Stdout is nil)."

%{  /* NOCONTEXT */

    if (@global(Stdout) == nil) {
        putchar(__intVal(__INST(asciivalue)));
        RETURN(self);
    }
%}.
    super print
!

printOn:aStream
    "print myself on aStream"

    aStream nextPut:self
!

printString
    "return a string to print me"

    ^ self asString
!

reconstructOn:aStream
    aStream nextPut:$$; nextPut:self

    "Created: 6.2.1996 / 09:46:12 / stefan"
    "Modified: 6.2.1996 / 10:57:35 / stefan"
!

storeOn:aStream
    "store myself on aStream"

    |special|

    (asciivalue between:33 and:127) ifFalse:[
	(self == Character space) ifTrue:[
	    special := 'space'
	] ifFalse:[
	    (self == Character cr) ifTrue:[
		special := 'cr'.
	    ] ifFalse:[
		(self == Character tab) ifTrue:[
		    special := 'tab'.
		] ifFalse:[
		    (self == Character esc) ifTrue:[
			special := 'esc'.
		    ]
		]
	    ]
	].
	special notNil ifTrue:[
	    aStream nextPutAll:'(Character '; nextPutAll:special; nextPut:$).
	    ^ self
	].
	aStream nextPutAll:'(Character value:16r'.
	asciivalue printOn:aStream base:16.
	aStream nextPut:$)
    ] ifTrue:[
	aStream nextPut:$$; nextPut:self
    ]

    "Modified: / 23.2.1996 / 23:27:32 / cg"
    "Modified: / 20.1.1998 / 14:10:46 / stefan"
! !

!Character methodsFor:'private-accessing'!

setAsciiValue:anInteger
    "very private - set the asciivalue / codePoint. 
     - use this only for characters with codes > 16rFF.
     DANGER alert: 
        funny things happen, if this is applied to
        one of the shared characters with codePoints 0..255."

    asciivalue := anInteger
! !

!Character methodsFor:'queries'!

bitsPerCharacter
    "return the number of bits I require for storage"

    asciivalue <= 16rFF ifTrue:[^ 8].
    asciivalue <= 16rFFFF ifTrue:[^ 16].
    ^ 32
!

stringSpecies
    "return the type of string that is needed to store me"

    asciivalue <= 16rFF ifTrue:[^ String].
    asciivalue <= 16rFFFF ifTrue:[^ Unicode16String].
    ^ Unicode32String
! !

!Character methodsFor:'testing'!

isCharacter
    "return true,  if the receiver is some kind of character"

    ^ true
!

isControlCharacter
    "return true if I am a control character (i.e. ascii value < 32)"

%{  /* NOCONTEXT */
#ifdef NON_ASCII       /* i.e. EBCDIC ;-) */
    not yet implemented - fails when compiled
#else
    REGISTER int val;

    val = __intVal(__INST(asciivalue));
    if (val < ' ') {
        RETURN ( true );
    }
#endif
    RETURN (false);
%}.
    ^ asciivalue < 16r20.

    "
     (Character value:1) isControlCharacter
     $a isControlCharacter
    "
!

isDigit
    "return true, if I am a digit (i.e. $0 .. $9)"

%{  /* NOCONTEXT */

    REGISTER int val;

    val = __intVal(__INST(asciivalue));
    if ((unsigned)(val - '0') <= ('9' - '0')) {
        RETURN ( true );
    }
    RETURN ( false );
%}.
    ^ asciivalue between:$0 asciiValue and:$9 asciiValue
!

isDigitRadix:r
    "return true, if I am a digit of a base r number"

    (asciivalue < $0 asciiValue) ifTrue:[^ false]. 
    (r > 10) ifTrue:[
        (asciivalue <= $9 asciiValue) ifTrue:[
            ^ true
        ].
        ((asciivalue - $a asciiValue) between:0 and:(r - 11)) ifTrue:[
            ^ true
        ].
        ^ (asciivalue - $A asciiValue) between:0 and:(r - 11)
    ].
    (asciivalue - $0 asciiValue) < r ifTrue:[^ true].
    ^ false
!

isEndOfLineCharacter
    "return true if I am a line delimitting character"

%{  /* NOCONTEXT */

    REGISTER int val;

    val = __intVal(__INST(asciivalue));
    if ((val == '\n')
     || (val == '\r')
     || (val == '\f')) {
        RETURN ( true );
    }
    RETURN (false);
%}.
    ^ asciivalue == 16r0A
      or:[asciivalue == 16r0D
      or:[asciivalue == 16r0C]]

!

isImmediate
    "return true if I am an immediate object
     i.e. I am represented in the pointer itself and
     no real object header/storage is used me.
     For VW compatibility, shared characters (i.e. in the range 0..255) 
     also return true here"

    ^ asciivalue <= 255

!

isLetter
    "return true, if I am a letter in the 'a'..'z' range.
     Use isNationalLetter, if you are interested in those."

%{  /* NOCONTEXT */

    REGISTER int val;

    val = __intVal(__INST(asciivalue));
    if ((unsigned)(val - 'a') <= ('z' - 'a')) {
        RETURN ( true );
    }
    if ((unsigned)(val - 'A') <= ('Z' - 'A')) {
        RETURN ( true );
    }
    RETURN ( false );
%}.
    ^ (asciivalue between:($a asciiValue) and:($z asciiValue))
      or:[(asciivalue between:($A asciiValue) and:($Z asciiValue))]
!

isLetterOrDigit
    "return true, if I am a letter (a..z or A..Z) or a digit (0..9)
     Use isNationalAlphaNumeric, if you are interested in those."

%{  /* NOCONTEXT */

    REGISTER int val;

    val = __intVal(__INST(asciivalue));
    if ((unsigned)(val - 'a') <= ('z' - 'a')) {
        RETURN ( true );
    }
    if ((unsigned)(val - 'A') <= ('Z' - 'A')) {
        RETURN ( true );
    }
    if ((unsigned)(val - '0') <= ('9' - '0')) {
        RETURN ( true );
    }
    RETURN ( false );
%}.
    ^ self isLetter or:[self isDigit]
!

isLowercase
    "return true, if I am a lower-case letter.
     This one does care for national characters."

%{  /* NOCONTEXT */

    REGISTER int val;
#define TRUE_IF_ODD(x)      ((x & 1) ? true : false)
#define TRUE_IF_EVEN(x)     ((x & 1) ? false : true)

    val = __intVal(__INST(asciivalue));
    switch (val >> 8) {
        case 0x00:
            if ((unsigned)(val - 'a') <= ('z' - 'a')) {
                RETURN ( true );
            }

            /* iso8859 puts national lower case characters at e0 .. ff */
            if ((val >= 0xE0) && (val <= 0xFF)) {
                if (val != 0xE7) {
                    RETURN(true);
                }
            }
            RETURN (false);

        case 0x01:
            if (val <= 0x0137) { RETURN (TRUE_IF_ODD(val)); } 
            if (val <= 0x0148) { RETURN (TRUE_IF_EVEN(val)); } 
            if (val <= 0x0178) { RETURN (TRUE_IF_ODD(val)); } 
            if (val <= 0x017E) { RETURN (TRUE_IF_EVEN(val)); } 
            if (val <= 0x017F) { RETURN (true); } 
            if (val < 0x01CD) {
                if (val == 0x01C6) { RETURN (true); }
                if (val == 0x01C9) { RETURN (true); }
                if (val == 0x01CC) { RETURN (true); }
                RETURN (false);
            } 
            if (val <= 0x01DC) { RETURN (TRUE_IF_EVEN(val)); } 
            if (val <= 0x01EF) { RETURN (TRUE_IF_ODD(val)); } 
            if (val == 0x01F0) { RETURN (true); }
            if (val == 0x01F1) { RETURN (false); }
            if (val == 0x01F2) { RETURN (false); }
            if (val == 0x01F3) { RETURN (true); }
            if (val <= 0x01F6) { RETURN (TRUE_IF_ODD(val)); } 
            if (val == 0x01F7) { RETURN (false); }
            RETURN (TRUE_IF_ODD(val));

        case 0x02:
            if (val <= 0x0233) { RETURN (TRUE_IF_ODD(val)); } 
            if (val <= 0x0236) { RETURN (true); } 
            if (val < 0x0250) { RETURN (false); } 
            if (val < 0x02B0) { RETURN (true); } 
            RETURN (false);


        case 0x03:
            if (val <= 0x03AB) { RETURN (false); } 
            if (val <= 0x03D1) { RETURN (true); } 
            if (val == 0x03D6) { RETURN (true); } 
            if (val <= 0x03D7) { RETURN (false); } 
            if (val <= 0x03EF) { RETURN (TRUE_IF_ODD(val)); } 
            if (val == 0x03F2) { RETURN (true); } 
            RETURN (false);

        case 0x04:
            if (val <= 0x042F) { RETURN (false); } 
            if (val <= 0x045F) { RETURN (true); } 
            if (val <= 0x0481) { RETURN (TRUE_IF_ODD(val)); } 
            if (val < 0x048A) { RETURN (false); } 
            RETURN (TRUE_IF_ODD(val));

        case 0x05:
            if (val <= 0x050F) { RETURN (TRUE_IF_ODD(val)); } 
            if (val < 0x0561) { RETURN (false); } 
            if (val <= 0x0487) { RETURN (true); } 
            RETURN (false);

        case 0xFF:
            if ((val >= 0xFF41) && (val <= 0xFF5A)) { RETURN (true); }
            RETURN (false);
    }
#undef TRUE_IF_ODD
#undef TRUE_IF_EVEN
    RETURN (false);
%}.
!

isPrintable
    "return true, if the receiver is a useful printable character
     (see fileBrowsers showFile:-method on how it can be used)"

    (asciivalue between:32 and:127) ifTrue:[^ true].
    asciivalue == 12 ifTrue:[^ true].   "/ FF
    asciivalue == 13 ifTrue:[^ true].   "/ CR
    asciivalue == 9 ifTrue:[^ true].    "/ TAB
    asciivalue == 10 ifTrue:[^ true].   "/ NL

    (asciivalue between:16rA0 and:16rBF) ifTrue:[^ true]. "/ ISO8859
    ^ self isNationalLetter

    "Modified: 7.8.1997 / 17:05:24 / cg"
!

isSeparator
    "return true if I am a space, cr, tab, nl, or newPage"

%{  /* NOCONTEXT */

    REGISTER int val;

    val = __intVal(__INST(asciivalue));
#ifndef NON_ASCII       /* i.e. EBCDIC ;-) */
    if (val <= ' ')
#endif
        if ((val == ' ')
         || (val == '\n') 
         || (val == '\t')
         || (val == '\r')
         || (val == '\f')) {
            RETURN ( true );
        }
    RETURN (false);
%}.                                  
    ^ (asciivalue == 16r20)
      or:[asciivalue == 16r0D
      or:[asciivalue == 16r0A
      or:[asciivalue == 16r09
      or:[asciivalue == 16r0C]]]]

!

isUppercase
    "return true, if I am an upper-case letter.
     This one does care for national characters."

%{  /* NOCONTEXT */
#define TRUE_IF_ODD(x)      ((x & 1) ? true : false)
#define TRUE_IF_EVEN(x)     ((x & 1) ? false : true)

    REGISTER int val;

    val = __intVal(__INST(asciivalue));

    switch (val >> 8) {
        case 0x00:
            if ((unsigned)(val - 'A') <= ('Z' - 'A')) {
                RETURN ( true );
            }
            /* iso8859 puts national upper case characters at c0 .. df */
            if ((val >= 0xC0) && (val <= 0xDF)) {
                if (val != 0xD7) {
                    RETURN(true);
                }
            }
            RETURN (false);

        case 0x01:
            if (val <= 0x0137) { RETURN (TRUE_IF_EVEN(val)); }
            if (val <= 0x0148) { RETURN (TRUE_IF_ODD(val)); } 
            if (val <= 0x0178) { RETURN (TRUE_IF_EVEN(val)); }
            if (val <= 0x017E) { RETURN (TRUE_IF_ODD(val)); } 
            if (val < 0x01CD) { 
                if (val == 0x01C4) { RETURN (true); }
                if (val == 0x01C5) { RETURN (true); }
                if (val == 0x01C7) { RETURN (true); }
                if (val == 0x01C8) { RETURN (true); }
                if (val == 0x01CA) { RETURN (true); }
                if (val == 0x01CB) { RETURN (true); }
                RETURN (false);         /* WRONG !!! */ 
            }
            if (val <= 0x01DC) { RETURN (TRUE_IF_ODD(val)); } 
            if (val <= 0x01EF) { RETURN (TRUE_IF_EVEN(val)); }
            if (val == 0x01F0) { RETURN (false); }
            if (val == 0x01F1) { RETURN (true); }
            if (val == 0x01F2) { RETURN (false); }
            if (val == 0x01F3) { RETURN (false); }
            if (val < 0x01F4) { RETURN (false); }
            if (val <= 0x01F7) { RETURN (TRUE_IF_ODD(val)); } 
            RETURN (TRUE_IF_EVEN(val));

        case 0x02:
            if (val <= 0x0233) { RETURN (TRUE_IF_EVEN(val)); }
            RETURN (false); 

        case 0x03:
            if (val < 0x0386) { RETURN (false); }
            if (val == 0x0387) { RETURN (false); }
            if (val <= 0x03AB) { RETURN (true); }
            if (val <= 0x03D1) { RETURN (false); }
            if (val <= 0x03D4) { RETURN (true); }
            if (val <= 0x03D7) { RETURN (false); }
            if (val <= 0x03EF) { RETURN (TRUE_IF_EVEN(val)); }
            RETURN (false); 

        case 0x04:
            if (val <= 0x042F) { RETURN (true); }
            if (val <= 0x045F) { RETURN (false); }
            if (val <= 0x0481) { RETURN (TRUE_IF_EVEN(val)); }
            if (val < 0x048A) { RETURN (false); }
            if (val == 0x04C0) { RETURN (false); }
            RETURN (TRUE_IF_EVEN(val));

        case 0x05:
            if (val <= 0x050F) { RETURN (TRUE_IF_EVEN(val)); }
            if (val < 0x0531) { RETURN (false); }
            if (val < 0x0556) { RETURN (true); }
            RETURN (false);

        case 0xFF:
            if ((val >= 0xFF21) && (val <= 0xFF3A)) { RETURN (true); }
            RETURN (false);
    }
    RETURN (false);

#undef TRUE_IF_ODD     
#undef TRUE_IF_EVEN    
%}
!

isVowel
    "return true, if I am a vowel (lower- or uppercase)"

    "/ I know the code is ugly;
    "/ better code is:
    "/     'aeiou' includes:self asLowercase
    "/ or:
    "/     'aeiouAEIOU' includes:self
    "/
    "/ until I have a smart compiler, I use the shorter (codewise):

    (self == $a) ifTrue:[^ true].
    (self == $e) ifTrue:[^ true].
    (self == $i) ifTrue:[^ true].
    (self == $o) ifTrue:[^ true].
    (self == $u) ifTrue:[^ true].
    (self == $A) ifTrue:[^ true].
    (self == $E) ifTrue:[^ true].
    (self == $I) ifTrue:[^ true].
    (self == $O) ifTrue:[^ true].
    (self == $U) ifTrue:[^ true].
    ^ false
! !

!Character methodsFor:'tracing'!

traceInto:aRequestor level:level from:referrer
    "double dispatch into tracer, passing my type implicitely in the selector"

    ^ aRequestor traceCharacter:self level:level from:referrer

! !

!Character class methodsFor:'documentation'!

version
    ^ '$Header: /cvs/stx/stx/libbasic/Character.st,v 1.103 2004-02-27 14:26:32 cg Exp $'
! !
